// Winsock.c -  Windows Socket Implemetation Example in C++

/* -----------------------------------------------------------------------------
	This is a code example using the Windows Socket feature in Windows 3.1 with C/C++.
	The example shows how to connect a client using a PC based Windows Application.
	The Windows	application uses the PCTCP WINSOCK.DLL (v1.1) libray.
	The code is well commented with steps to follow the creation of the application as
	well as comments for common pitfalls and important lines of code that affect the
	performance of your application.
	The Client and Server uses the services files for the port access, "example/tcp"
 ------------------------------------------------------------------------------ */

#include <stdio.h>                                   	
#include <stdlib.h>
#include <dir.h>
#include <array.h>
#include <abstarry.h>
#include <string.h>
#include <mem.h>
#include <owl.h>
#include <windobj.h>
#include <dialog.h>
#include "ctype.h" 	
#include <inputdia.h>
#include <filedial.h>
#include "sockfile.h"
#include "edit.h"
#include "button.h"
#include "static.h"
extern "C" {
    #include "winsock.h"
}

#define DESIRED_WINSOCK_VERSION		0x0101	    // Version 1.1 of WINSOCK.DLL
#define MINIMUM_WINSOCK_VERSION		0x0001		// Version 1.0 of WINSOCK.DLL
#define MINIMUM_SOCKETS_REQUIRED 	6
#define MAX_BUF_LEN					60
#define	TRANSCEIVE_BUFFER_LEN		200
#define TRUE						1
#define FALSE						0
#define CLOSED						0
#define OPEN						1
#define	CLIENT						1
#define SERVER						0

	struct	sockaddr_in	cli_addr;  				// Server address struct
	struct	sockaddr_in srv_addr;  				// Client address struct
	struct	hostent		*hp;
    struct	servent		*sp;
	WSADATA				wsaData;
	LPSERVENT			srv_info;
	LPHOSTENT			host_info;
	SOCKET				srv_sock;
	SOCKET				cli_sock;
	SOCKET				new_sock;
    SOCKET				s;
	char				buf[MAX_BUF_LEN];
	char				rec_buffer[TRANSCEIVE_BUFFER_LEN];
	char				socket_id = SERVER;
	HBITMAP				hLogo;
	BITMAP				bm;

	LINGER				lingerinfo;

	struct TTransferStruct {
		char 	text_buffer[TRANSCEIVE_BUFFER_LEN];
	};

	static char	status1[] = "WSAStartup completed, Winsock.DLL Version 1.10";
	static char	status2[] = "Socket Created and Bind to address in services";
	static char status3[] = "Client in Asynchronous mode : FD_WRITE or FD_CLOSE signal alerts";
	static char status4[] = "Server in LISTEN mode : FD_ACCEPT signal alerts";
	static char status5[] = "Server Accepts client request, Creates new socket and opens TCP/IP connection";
	static char status6[] = "Server in Asynchronous mode : FD_READ or FD_CLOSE signal alerts";

class TSendDlg : public TDialog
{
public:
  int		text_bytes;
  char 		send_buffer[TRANSCEIVE_BUFFER_LEN];
  TSendDlg(PTWindowsObject AParent, LPSTR name);
  virtual void CMSendData(RTMessage Msg)     	= [ID_FIRST + ID_SEND];
  virtual void CMCancelSend(RTMessage Msg)    	= [ID_FIRST + ID_CANCEL];
  virtual void CMAsyncWrite(RTMessage Msg)    	= [WM_FIRST + SOCKET_MESSAGE1];
};

_CLASSDEF(TMyWindow)
class TMyWindow : public TWindow
{
public:
  char FileName[MAXPATH];
  TMyWindow(PTWindowsObject AParent, LPSTR ATitle);
  ~TMyWindow();
  TTransferStruct TransferStruct;
  virtual void Paint(HDC PaintDC, PAINTSTRUCT _FAR &Paintinfo);
  virtual BOOL CanClose();
  virtual void CMFileNew(RTMessage Msg)     	= [CM_FIRST + CM_FILENEW];
  virtual void CMFileOpen(RTMessage Msg)    	= [CM_FIRST + CM_FILEOPEN];
  virtual void CMFileSave(RTMessage Msg)    	= [CM_FIRST + CM_FILESAVE];
  virtual void CMFileSaveAs(RTMessage Msg)  	= [CM_FIRST + CM_FILESAVEAS];
  virtual void CMClientOpen(RTMessage Msg) 	= [CM_FIRST + CM_CLIENTOPEN];
  virtual void CMServerOpen(RTMessage Msg) 	= [CM_FIRST + CM_SERVEROPEN];
  virtual void CMDataTransRec(RTMessage Msg) 	= [CM_FIRST + CM_DATATRANSREC];
  virtual void CMSessionClose(RTMessage Msg)	= [CM_FIRST + CM_SESSIONCLOSE];
  virtual void CMHelp(RTMessage Msg)        	= [CM_FIRST + CM_HELP];
  virtual void CMAbout(RTMessage Msg)        	= [CM_FIRST + CM_ABOUT];
  virtual void CMAsyncRead(RTMessage Msg)    	= [WM_FIRST + SOCKET_MESSAGE];
};

class TMyApp : public TApplication
{
public:
  TMyApp(LPSTR AName, HINSTANCE hInstance, HINSTANCE hPrevInstance,
    LPSTR lpCmdLine, int nCmdShow)
    : TApplication(AName, hInstance, hPrevInstance, lpCmdLine, nCmdShow) {};
  virtual void InitMainWindow();
  virtual void InitInstance();
};

// *************************************************
// TSendDlg Implementation
// *************************************************
TSendDlg::TSendDlg(PTWindowsObject AParent, LPSTR name)
  : TDialog(AParent, name)
{
	new TEdit(this, ID_EDITTEXT,
		sizeof(((TMyWindow *)Parent)->TransferStruct.text_buffer));
	new TButton(this, ID_SEND);
	text_bytes = 0;
}

void TSendDlg::CMSendData(RTMessage Msg)
{
	//CloseWindow();
	if (Msg.LP.Hi == BN_CLICKED)
	{
		if ( (text_bytes = GetDlgItemText(HWindow, ID_EDITTEXT,
										  send_buffer, TRANSCEIVE_BUFFER_LEN)) == 0)
		{
			MessageBox(HWindow, "No TEXT to TRANSMIT", "Client Operation", MB_OK);
			return;
		}
		else
		{
			sprintf(buf,"Bytes tranferred to send_buffer = %d\n send_buffer = %s",
					text_bytes, send_buffer);
			MessageBox(HWindow,buf,"Client Operation",MB_ICONSTOP | MB_OK);

			SendMessage(HWindow, SOCKET_MESSAGE1, 0, MAKELONG(FD_WRITE,0));
			//MessageBox(HWindow, "Message FD_WRITE successfully sent", "Client Operation", MB_OK);
		};
	}
}

void TSendDlg::CMCancelSend(RTMessage)
{
	CloseWindow();
}

void TSendDlg::CMAsyncWrite(RTMessage Msg)
{
	int			bytes_sent;
  	int			buffer_length;
	int			total_bytes_sent = 0;

	switch( Msg.LP.Lo ) {

	case FD_WRITE:

		//
		// DO NOT Send data on the first FD_WRITE signal as it is invoked by FD_CONNECT
		//

		if (WSAGetLastError() == WSAENETDOWN)
		{
			MessageBox(HWindow,"Network Down or Bad Connection","Windows Sockets Error",
					   MB_OK|MB_ICONQUESTION);
			break;
		}

		//
		// Write text in edit box (loop until all data in buffer is send
		//

		if (socket_id == CLIENT)
			s = cli_sock;
		else
			s = new_sock;

		buffer_length = sizeof(send_buffer);
		do
  		{
			bytes_sent = send(s, (char FAR *)&send_buffer, buffer_length, 0);

			if ( bytes_sent == SOCKET_ERROR ) {
				sprintf(buf,"Could not send Data. /nError : %d",WSAGetLastError());
				MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
				return;
			}
			total_bytes_sent += bytes_sent;
			buffer_length -= bytes_sent;
		} while ((buffer_length > 0) || (WSAGetLastError() == WSAEWOULDBLOCK));

  		sprintf(buf,"Bytes send : %d, BUFFER = %s" ,total_bytes_sent, send_buffer);
		MessageBox(HWindow, buf, "Client Operation", MB_OK);
		break;

	} // End of case
}

// *************************************************
// TMyWindow Implementation
// *************************************************
TMyWindow::TMyWindow(PTWindowsObject AParent, LPSTR ATitle)
  : TWindow(AParent, ATitle)
{
  Attr.Style |= WS_VISIBLE | WS_MAXIMIZE;
  AssignMenu("COMMANDS");
  memset(&TransferStruct, 0x0, sizeof TransferStruct);
}

TMyWindow::~TMyWindow()
{
}

BOOL TMyWindow::CanClose()
{
  WSACleanup();
  return MessageBox(HWindow, "Do you want to Quit and Save?",
	"Drawing has changed", MB_YESNO | MB_ICONQUESTION) == IDYES;
}

void TMyWindow::Paint(HDC PaintDC, PAINTSTRUCT _FAR & Paintinfo)
{
	HDC	hMemDC;

	hLogo  = LoadBitmap(GetApplication()->hInstance,"BITMAP_2");
	hMemDC = CreateCompatibleDC(Paintinfo.hdc);
	SelectObject(hMemDC, hLogo);
	//BitBlt(PaintDC, 30, 50, bm.bmWidth, bm.bmHeight, hMemDC, 0, 0, SRCCOPY);
	BitBlt(PaintDC, 0, 0, 640, 480, hMemDC, 0, 0, SRCCOPY);
	DeleteDC(hMemDC);
    
}

void TMyWindow::CMFileNew(RTMessage)
{
  MessageBox(HWindow, "Feature not implemented", "File Save", MB_OK);
}

void TMyWindow::CMFileOpen(RTMessage)
{
  if ( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN,
    	strcpy(FileName, "*.PTS"))) == IDOK )
	 MessageBox(HWindow, FileName, "Open the file:", MB_OK);
}

void TMyWindow::CMFileSave(RTMessage)
{
  MessageBox(HWindow, "Feature not implemented", "File Save", MB_OK);
}


void TMyWindow::CMClientOpen(RTMessage)
{
  char 			Newhost[] = "196.6.126.20";
  int			serr;
  int 			addr_len;
  struct in_addr	inet;
  HDC			cli_hdc;

  socket_id = CLIENT;

  //
  // Initialise Windows Sockets
  //
  
  serr = WSAStartup(DESIRED_WINSOCK_VERSION,&wsaData);
  if (serr != 0)
  { 
	sprintf(buf,"Cannot Initialise socket library",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_ICONSTOP | MB_OK);
	return;
  }

  if( wsaData.wVersion < MINIMUM_WINSOCK_VERSION )
  {
	sprintf(buf,"Windows Sockets version %02X.%02X, Need at Least %02X.%02X",
		LOBYTE(wsaData.wVersion),
		HIBYTE(wsaData.wVersion),
		LOBYTE(MINIMUM_WINSOCK_VERSION),
		HIBYTE(MINIMUM_WINSOCK_VERSION));
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_ICONSTOP | MB_OK);
  }
  
  if (wsaData.iMaxSockets < MINIMUM_SOCKETS_REQUIRED ) {
  	sprintf(buf,"Application requires a minimum of %d supported sockets: /nError : %d",
		MINIMUM_SOCKETS_REQUIRED, WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_ICONSTOP | MB_OK);
  }

  cli_hdc = GetDC(HWindow);
  if (TextOut(cli_hdc, 0, 0, (LPCSTR)&status1, strlen(status1)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
	ReleaseDC(HWindow,cli_hdc);

  //
  // Create a stream socket
  //
  	
  cli_sock = socket(AF_INET, SOCK_STREAM, 0);
  if ( cli_sock == INVALID_SOCKET )
  {
	sprintf(buf,"Could not Create Socket,error : %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  }

  serr = ( cli_sock == INVALID_SOCKET ) ? WSAGetLastError() : 0;
  if( serr == 0 )
  {

		//
        //  Bind an address to the socket.
		//

	cli_addr.sin_family      = AF_INET;
	cli_addr.sin_addr.s_addr = INADDR_ANY;
	cli_addr.sin_port        = 0;

	if( bind( cli_sock, (LPSOCKADDR)&cli_addr, sizeof(cli_addr) ) == SOCKET_ERROR )
	{
		sprintf(buf,"Could not Bind Socket. /nError : %d",WSAGetLastError());
		MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
	}
  }

  cli_hdc = GetDC(HWindow);
  //str_length = LoadString(sth_startup,1,str_buffer,100);
  if (TextOut(cli_hdc, 0, 20, (LPCSTR)&status2, strlen(status2)) == 0)
  //if (TextOut(cli_hdc, 0, 20, MAKEINTRESOURCE(sth_socket), strlen(MAKEINTRESOURCE(sth_socket))) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
	ReleaseDC(HWindow,cli_hdc);

  //
  // Get Newhost and determine if dotted IP address
  //

  if (GetApplication()->ExecDialog(new TInputDialog(this, "Enter HOSTNAME or IP Address", "",
		Newhost, sizeof(Newhost))) == IDOK){
  }
  else
  {
		WSACleanup();
		return;
  } 

  //
  // First check to see if it is a dotted IP address
  //
  inet  = inet_addr(Newhost);
 // if (inet.laddr == INADDR_NONE)
  {
//  		hp = gethostbyname(Newhost);
  }
//  else
  {
		hp = gethostbyaddr((char far *)&inet,sizeof(inet),PF_INET);
  };

  if (hp == NULL) {
  		sprintf(buf,"Could not get hostname or host address, Error : %d",WSAGetLastError());
		MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  }

  //
  // Get remote port ID to connect to for FTP service
  //

  srv_info = getservbyname("example","tcp");
  if ( srv_info == NULL ) {
	   sprintf(buf,"Could not resolve FTP service port, Error : %d",WSAGetLastError());
	   MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  } 

  srv_addr.sin_family = AF_INET;
  srv_addr.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
  srv_addr.sin_port = srv_info->s_port;

  //
  // Connect client socket to FTP server at address SERVER
  //

  		if ( connect(cli_sock,(LPSOCKADDR)&srv_addr, sizeof(srv_addr) ) == SOCKET_ERROR ) {
	   		sprintf(buf,"Could not Connect Socket. Error : %d",WSAGetLastError());
	   		MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  		}

  		//
  		// Identify the Socket Information to what we are connected
  		//

  		addr_len = sizeof(struct sockaddr_in);
 	 	if (getsockname(cli_sock, (LPSOCKADDR)&cli_addr, &addr_len ) == SOCKET_ERROR ) {
  			sprintf(buf,"Cannot Read Socket addres/name. Error : %d",WSAGetLastError());
			MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  		}
  		sprintf(buf,"Connected to local port : %u", ntohs(cli_addr.sin_port));
		MessageBox(HWindow,buf,"Client Operation",MB_OK);

  //
  // Gracefully close the client (with 1 sec linger) when all data is transmitted
  //

  lingerinfo.l_onoff = 1;
  lingerinfo.l_linger = 1;
  if (setsockopt(cli_sock, SOL_SOCKET, SO_LINGER, (char *)&lingerinfo, sizeof(lingerinfo)) == SOCKET_ERROR )
  {
	MessageBox(HWindow, "Error in setting socket option", "Client Operation", MB_OK);
  }

  //
  // Set up Async select on FD_WRITE and FD_CLOSE 
  //

  if (WSAAsyncSelect(cli_sock, HWindow, SOCKET_MESSAGE, FD_WRITE|FD_READ|FD_CLOSE) == SOCKET_ERROR)
  {
	sprintf(buf,"WSAAsyncSelect failure on [cli_sock]. Error : %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
    return;
  }

  cli_hdc = GetDC(HWindow);
  if (TextOut(cli_hdc, 0, 40, (LPCSTR)&status3, strlen(status3)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
	ReleaseDC(HWindow,cli_hdc);

}


void TMyWindow::CMServerOpen(RTMessage)
{

  int			serr;
  int			addr_len;
  HDC			hdc;

  socket_id = SERVER;

  serr = WSAStartup(DESIRED_WINSOCK_VERSION,&wsaData);
  if (serr != 0)
  { 
	sprintf(buf,"Cannot Initialise socket library",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_ICONSTOP | MB_OK);
	return;
  }

  if( wsaData.wVersion < MINIMUM_WINSOCK_VERSION )
  {
	sprintf(buf,"Windows Sockets version %02X.%02X, Need at Least %02X.%02X",
		LOBYTE(wsaData.wVersion),
		HIBYTE(wsaData.wVersion),
		LOBYTE(MINIMUM_WINSOCK_VERSION),
		HIBYTE(MINIMUM_WINSOCK_VERSION));
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_ICONSTOP | MB_OK);
  }
  
  if (wsaData.iMaxSockets < MINIMUM_SOCKETS_REQUIRED )
  {
  	sprintf(buf,"Application requires a minimum of %d supported sockets: /nError : %d",
		MINIMUM_SOCKETS_REQUIRED, WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_ICONSTOP | MB_OK);
  }

  hdc = GetDC(HWindow);
  if (TextOut(hdc, 0, 0, (LPCSTR)&status1, strlen(status1)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
  ReleaseDC(HWindow,hdc);

  //
  // Create a stream socket
  //
  	
  srv_sock = socket(AF_INET, SOCK_STREAM, 0);
  if ( srv_sock == INVALID_SOCKET )
  {
	sprintf(buf,"Could not Create Socket,error : %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  }

  serr = ( srv_sock == INVALID_SOCKET ) ? WSAGetLastError() : 0;
  if( serr == 0 )
  {

	srv_info = getservbyname("example","tcp");
  	if ( srv_info == NULL ) {
	   sprintf(buf,"Could not resolve FTP service port, Error : %d",WSAGetLastError());
	   MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  	} 


	//
    //  Bind socket to the appropriate port and interface(INADDR_ANY).
    //
	srv_addr.sin_family      = AF_INET;
	srv_addr.sin_addr.s_addr = INADDR_ANY;
	srv_addr.sin_port        = srv_info->s_port;

	if( bind( srv_sock, (LPSOCKADDR)&srv_addr, sizeof(srv_addr) ) == SOCKET_ERROR )
	{
		sprintf(buf,"Could not Bind Socket. /nError : %d",WSAGetLastError());
		MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
        return;
	}
  }

  hdc = GetDC(HWindow);
  if (TextOut(hdc, 0, 20, (LPCSTR)&status2, strlen(status2)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
  ReleaseDC(HWindow,hdc);

  //
  // Listen for incoming connection
  //

  if (listen(srv_sock,1) == SOCKET_ERROR)
  {
	sprintf(buf,"Cannot set up listen on Socket. Error : %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
    return;
  }

  hdc = GetDC(HWindow);
  if (TextOut(hdc, 0, 40, (LPCSTR)&status4, strlen(status4)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
	ReleaseDC(HWindow,hdc);

  //
  // Accept and service incoming connection
  //

  addr_len = sizeof(struct sockaddr_in);
  new_sock = accept(srv_sock,(LPSOCKADDR)&cli_addr,&addr_len);
  if (cli_sock == INVALID_SOCKET)
  {
	sprintf(buf,"Cannot accept incoming connection on Socket. Error : %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
    return;
  }

  hdc = GetDC(HWindow);
  if (TextOut(hdc, 0, 60, (LPCSTR)&status5, strlen(status5)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
  ReleaseDC(HWindow,hdc);

  //
  // Set up Async select on FD_READ and FD_CLOSE
  //

  if (WSAAsyncSelect(new_sock, HWindow, SOCKET_MESSAGE, FD_READ | FD_CLOSE) == SOCKET_ERROR)
  {
	sprintf(buf,"WSAAsyncSelect failure on [srv_sock]. Error : %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
    return;
  }

  hdc = GetDC(HWindow);
  if (TextOut(hdc, 0, 80, (LPCSTR)&status6, strlen(status6)) == 0)
  {
	MessageBox(HWindow, "Problem with TextOut Routine", "Status Message", MB_OK);
  }
  ReleaseDC(HWindow,hdc);

}

void TMyWindow::CMAsyncRead(RTMessage Msg)
{

	int			bytes_received;
  	int			total_bytes_received = 0;
	int 		rec_buffer_length;

	//HDC			hdc;

	switch( Msg.LP.Lo ) {

	case FD_READ:

		//
		// Data arrived. Receive it
		//

		if (socket_id == CLIENT)
			s = cli_sock;
		else
			s = new_sock;

		rec_buffer_length = sizeof(rec_buffer);
		memset(rec_buffer,' ',rec_buffer_length-1);

		do
  		{	
			bytes_received = recv(s, rec_buffer, rec_buffer_length, 0);
			if (WSAGetLastError() == WSAECONNABORTED)
			{
				MessageBox(HWindow,"Connection already closed","Windows Sockets Error",MB_OK);
				break;
			}

			if (WSAGetLastError() == WSAEWOULDBLOCK)
			{
				MessageBox(HWindow,"Data already received","Windows Sockets Error",MB_OK);
				break;
			}

			if (bytes_received == SOCKET_ERROR)
			{
				sprintf(buf,"Error while receiving data: %d",WSAGetLastError());
				MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
				break;
			}

			total_bytes_received += bytes_received;
			rec_buffer_length -= bytes_received;

		} while (rec_buffer_length > 0);

  		sprintf(buf,"\tBytes Received : %d\n\nBUFFER = %s  ", total_bytes_received, rec_buffer);
		MessageBox(HWindow, buf, "Data Receive Section", MB_OKCANCEL);
		break;

	case FD_CLOSE:

		//
  		// Close Server socket if state is CLOSED
  		//
		if (socket_id == CLIENT)
		{
			if (closesocket(cli_sock) == SOCKET_ERROR)
				MessageBox(HWindow, "Error during Shutdown", "Client Operation", MB_OK);
			MessageBox(HWindow, "Client Socket Gracefully closed", "Client Operation", MB_OK);
		}
		else
		{
			if (closesocket(new_sock) == SOCKET_ERROR)
			MessageBox(HWindow, "Error during Shutdown", "Server Operation", MB_OK);

			if (WSAAsyncSelect(new_sock, HWindow, SOCKET_MESSAGE, 0) == SOCKET_ERROR)
  			{
				sprintf(buf,"WSAAsyncSelect failure on [Server sock]. Error : %d",WSAGetLastError());
				MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
    			return;
			}
			MessageBox(HWindow, "Server Socket Gracefully closed", "Server Operation", MB_OK);
		};
		break;

	} //End of Switch

}

void TMyWindow::CMDataTransRec(RTMessage)
{
   if (GetModule()->ExecDialog(new TSendDlg(this,"SD_STATUS")) == ID_SEND)
   {
	   MessageBox(HWindow, "Send in Progress", "Client Operation", MB_OK);
   }
}

void TMyWindow::CMSessionClose(RTMessage)
{
  MessageBox(HWindow, "Press Enter to Close Socket", "Client Operation", MB_OK);

  //
  // Post FD_CLOSE to message queue
  //

  SendMessage(HWindow, SOCKET_MESSAGE, 0, MAKELONG(FD_CLOSE,0));
  MessageBox(HWindow, "Message FD_CLOSE successfully sent", "Client Operation", MB_OK);

  //
  // CleanUp the Opened Socket
  //

  if (WSACleanup() == SOCKET_ERROR)
  {
	sprintf(buf,"Error while executing WSACleanup: %d",WSAGetLastError());
	MessageBox(HWindow,buf,"Windows Sockets Error",MB_OK);
  }
  else
	MessageBox(HWindow, "WSACleanup successful", "Client/Server Session Close", MB_OK);
}

void TMyWindow::CMFileSaveAs(RTMessage)
{
  if ( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILESAVE,
    FileName)) == IDOK )
      MessageBox(HWindow, FileName, "Save the file:", MB_OK);
}


void TMyWindow::CMHelp(RTMessage)
{
  //
  // Pop Up SOCKFILE.HLP window from the HELP MENU
  //

  WinHelp ( HWindow, "SOCKFILE.HLP", HELP_CONTENTS, 0L );
}

void TMyWindow::CMAbout(RTMessage)
{
  //
  // Pop Up About window
  //

  PTWindow	HelpWindow;

  HelpWindow = new TWindow(this, "About System");
  HelpWindow->Attr.Style |= WS_POPUPWINDOW | WS_CAPTION;
  HelpWindow->Attr.X = 100;
  HelpWindow->Attr.Y = 100;
  HelpWindow->Attr.W = 300;
  HelpWindow->Attr.H = 300;
  GetApplication()->MakeWindow(HelpWindow);
}


void TMyApp::InitMainWindow()
{
  MainWindow = new TMyWindow(NULL, "Windows Client / Server Socket Program");
}


void TMyApp::InitInstance()
{
	TApplication::InitInstance();
	HAccTable = LoadAccelerators(hInstance,"ACCELERATORS_1");
}


// Main Program
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
  LPSTR lpCmdLine, int nCmdShow)
{
  TMyApp MyApp("Windows Socket Program", hInstance, hPrevInstance,
               lpCmdLine, nCmdShow);
  MyApp.Run();
  return MyApp.Status;
}
