//*******************************************************************
//  Program to convert the original WS_FTP ini file WINSOCK.INI to 
//  a format understood by the modified WS_FTP program.
//*******************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>
#include <windows.h>

#include "winsock.h"
#include "ws_ftp.h"

char szOldFile[80]="winsock.ini";
char szNewFile[80]="ws_ftp.ini";
char **lpMemBuf;

int nLines, nCfg, nCfgNum;
int nTempDir=0, nConfig=0, nRetain=0, nSection=0;

LPSTR lpExt[6] = { ".TXT", ".ZIP", ".BMP", ".GIF", ".WRI", ".LST" };
LPSTR lpViewer[6] = { "notepad", "wizunzip", "wingif", "wingif", "write", "gander" };
LPSTR lpTyp[6] = { "A", "I", "I", "I", "I", "A" };

int nOldHost[15] = { 0, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096 };
int nNewHost[15] = { 0, 1, 2, 3, 4,  5,  6,  7,   8,   9,  10,   11,   12,   14 };

LPHOSTCONFIG lpHost;

main (int argc, char *argv[])
{
  FILE *fp;
  int nI;
  
  if (argc>1)
  {
    if (argv[1]!=NULL)
    {
      strcpy (szOldFile, argv[1]);                      // Old Ini file
      if (argv[2]!=NULL) strcpy (szNewFile, argv[2]);   // New Ini File
    }
  }

  lpMemBuf = (char **) malloc (200 * sizeof (char *));
  for (nI=0; nI<200; nI++) lpMemBuf[nI] = malloc (100);
  for (nI=0; nI<200; nI++) memset (lpMemBuf[nI], '\0', 100);

  //*********************************************************
  // Read in the old file
  //*********************************************************
  fp = fopen (szOldFile, "rt");
  nI=0;
  while ((nI<200) && (fgets (lpMemBuf[nI], 99, fp)!=NULL))
  {
    strtok (lpMemBuf[nI++], "\n");
  }
  fclose (fp);
  nLines = nI;


  //*********************************************************
  //  Find out the total number of configurations, and 
  //  allocate sufficient memory to hold them all.
  //*********************************************************
  for (nI=0, nCfg=0; nI<nLines; nI++)
  {
    if (strnicmp (lpMemBuf[nI], "HST", 3)==0) nCfg = __max (nCfg, atoi (lpMemBuf[nI]+3));
    else if (strnicmp (lpMemBuf[nI], "HOST", 4)==0) nCfg = __max (nCfg, atoi (lpMemBuf[nI]+4));
  }
  nCfg++;
  lpHost = (LPHOSTCONFIG) malloc (nCfg * sizeof (HOSTCONFIG));
  memset (lpHost, '\0', nCfg * sizeof (HOSTCONFIG));
  
  //******************************************************************
  // Extract the configuration info for all the configurations
  //******************************************************************
  for (nI=0; nI<nLines; nI++)
  {
    if (strnicmp (lpMemBuf[nI], "CFG", 3)==0) ConvertRecord (nI, 0, 3);
    if (strnicmp (lpMemBuf[nI], "HST", 3)==0) ConvertRecord (nI, 1, 3);
    if (strnicmp (lpMemBuf[nI], "UID", 3)==0) ConvertRecord (nI, 2, 3);
    if (strnicmp (lpMemBuf[nI], "PWD", 3)==0) ConvertRecord (nI, 3, 3);
    if (strnicmp (lpMemBuf[nI], "DIR", 3)==0) ConvertRecord (nI, 4, 3);
  }

  //******************************************************************
  //  Create Configuration Names
  //  Extract the Host Type & TimeOut info
  //******************************************************************
  for (nI=0; nI<nCfg; nI++)
  {
    char *lp;
    int nTmp, nJ;
    
    lp = lpHost[nI].szHostName;
    lp = strchr (lp, ' ');
    if (lp!=NULL) 
    {
      *lp++ = '\0'; while (*lp==' ') lp++;
      sscanf (lp, "%d %d", &nTmp, &lpHost[nI].nTimeOut);
      for (nJ=0; nJ<15; nJ++) if (nOldHost[nJ]==nTmp) { nTmp = nNewHost[nJ]; break; };
      lpHost[nI].nType = __min (nTmp, 14);
    }
    if (lpHost[nI].szConfig[0]=='\0') strcpy (lpHost[nI].szConfig, lpHost[nI].szHostName);
  }

  for (nI=0; nI<nLines; nI++)
  {
    if (strnicmp (lpMemBuf[nI], "[WS_FTP]", 8)==0) free (lpMemBuf[nI]), lpMemBuf[nI]=NULL;
    if (strnicmp (lpMemBuf[nI], "TempDir", 7)==0) nTempDir=1;
    if (strnicmp (lpMemBuf[nI], "ConfigNum", 9)==0) nConfig=1;
    if (strnicmp (lpMemBuf[nI], "Retain", 7)==0) nRetain=1;
  }
  
  //******************************************************************
  //  Create the output file
  //******************************************************************
  fp = fopen (szNewFile, "wt");

  //******************************************************************
  // Write out the [WS_FTP] section information
  //******************************************************************
  fprintf (fp, "[WS_FTP]\n");
  for (nI=0; nI<nLines; nI++)
  {
    if (lpMemBuf[nI]!=NULL) fprintf (fp, "%s\n", lpMemBuf[nI]);
  }
  if (!nTempDir) fprintf (fp, "TempDir=C:\\\n");
  if (!nConfig) fprintf (fp, "ConfigNum=20\n");
  if (!nRetain) fprintf (fp, "Retain=0\n");
  fprintf (fp, "ViewerCount=6\n");
  
  //*****************************************************************
  //  Write out the configuration list
  //*****************************************************************
  for (nI=0, nCfgNum=0; nI<nCfg; nI++)
  {
    if (lpHost[nI].szConfig[0]!='\0')
    {
      if (lpHost[nI].szHostName[0]!='\0')
        fprintf (fp, "CFG%d=%s\n", nCfgNum++, lpHost[nI].szConfig);
    }
  }
  fprintf (fp, "\n");

  fprintf (fp, "[WS_FTP:Viewer]\n");
  for (nI=0; nI<6; nI++) 
  {
    fprintf (fp, "VIEWER%d=%s,%s,%s\n", nI+1, lpTyp[nI], lpExt[nI], lpViewer[nI]);
  }
  fprintf (fp, "\n");

  //*****************************************************************
  //  Now write out the individual configuration info
  //*****************************************************************
  for (nI=0, nCfgNum=0; nI<nCfg; nI++)
  {
    if (lpHost[nI].szConfig[0]!='\0')
    {
      fprintf (fp, "[WS_FTP:%s]\n", lpHost[nI].szConfig);
      if (lpHost[nI].szHostName[0]!='\0') fprintf (fp, "HostName=%s\n", lpHost[nI].szHostName, lpHost[nI].nType);
      if (lpHost[nI].szUserID[0]!='\0')   fprintf (fp, "UserID=%s\n", lpHost[nI].szUserID);
      if (lpHost[nI].szPassword[0]!='\0') fprintf (fp, "Password=%s\n", lpHost[nI].szPassword);
      if (lpHost[nI].szInitDir[0]!='\0')  fprintf (fp, "InitDir=%s\n", lpHost[nI].szInitDir);
      fprintf (fp, "TimeOut=%d\n", __max (lpHost[nI].nTimeOut, 65));
      fprintf (fp, "HostType=%d\n", lpHost[nI].nType);
      fprintf (fp, "\n");
    }
  }

  fclose (fp);
  
  for (nI=0; nI<200; nI++) if (lpMemBuf[nI]!=NULL) free (lpMemBuf[nI]);
  free (lpMemBuf);
  free (lpHost);
}

//***********************************************************************
//  Convert one line of the configuration to the appropriate record
//***********************************************************************
ConvertRecord (int nIndex, int nTyp, int nLen)
{
  char *lp, *lpBuf;
  int nCfgNum;
  
  lpBuf = lpMemBuf[nIndex];
  lp = strchr (lpBuf, '=');
  if (lp!=NULL)
  {
    *lp++ = '\0';
    while (*lp==' ') lp++;
    nCfgNum = atoi (lpBuf+3);
    switch (nTyp)
    {
      case 0: strcpy (lpHost[nCfgNum].szConfig,   lp); break;  // Fit into Config Entry
      case 1: strcpy (lpHost[nCfgNum].szHostName, lp); break;  // Fit into HostName Entry
      case 2: strcpy (lpHost[nCfgNum].szUserID,   lp); break;  // Fit into UserID Entry
      case 3: strcpy (lpHost[nCfgNum].szPassword, lp); break;  // Fit into Password Entry
      case 4: strcpy (lpHost[nCfgNum].szInitDir,  lp); break;  // Fit into InitDir Entry
    }
    
    //*********************************************************
    //  Now that Config Information has been extracted,
    //  dispose of this memory buffer
    //*********************************************************
    free (lpMemBuf[nIndex]);
    lpMemBuf[nIndex] = NULL;
  }
}
